import { DEG2RAD, RAD2DEG, TAU } from 'ootk-core';
/**
 * Converts Azimuth and Elevation to U and V.
 * Azimuth is the angle off of boresight in the horizontal plane.
 * Elevation is the angle off of boresight in the vertical plane.
 * Cone half angle is the angle of the cone of the radar max field of view.
 * @param az - Azimuth in radians
 * @param el - Elevation in radians
 * @param coneHalfAngle - Cone half angle in radians
 * @returns U and V in radians
 */
export function azel2uv(az, el, coneHalfAngle) {
    if (az > coneHalfAngle && az < coneHalfAngle) {
        throw new RangeError(`Azimuth is out of bounds: ${az}`);
    }
    if (el > coneHalfAngle && el < coneHalfAngle) {
        throw new RangeError(`Elevation is out of bounds: ${el}`);
    }
    const alpha = (az / (coneHalfAngle * RAD2DEG)) * 90;
    const beta = (el / (coneHalfAngle * RAD2DEG)) * 90;
    const u = Math.sin(alpha);
    let v = -Math.sin(beta);
    v = Object.is(v, -0) ? 0 : v;
    return { u, v };
}
/**
 * Determine azimuth and elevation off of boresight based on sensor orientation and RAE.
 * @param rae Range, Azimuth, Elevation
 * @param sensor Radar sensor object
 * @param face Face number of the sensor
 * @param maxSensorAz Maximum sensor azimuth
 * @returns Azimuth and Elevation off of boresight
 */
export function rae2raeOffBoresight(rae, sensor, face, maxSensorAz) {
    let az = (rae.az * DEG2RAD);
    let el = (rae.el * DEG2RAD);
    // Correct azimuth for sensor orientation.
    az = az > maxSensorAz * DEG2RAD ? (az - TAU) : az;
    az = (az - sensor.boresightAz[face]);
    el = (el - sensor.boresightEl[face]);
    return { az, el };
}
/**
 * Converts Range Az El to Range U V.
 * @param rae Range, Azimuth, Elevation
 * @param sensor Radar sensor object
 * @param face Face number of the sensor
 * @param maxSensorAz Maximum sensor azimuth
 * @returns Range, U, V
 */
export function rae2ruv(rae, sensor, face, maxSensorAz) {
    const { az, el } = rae2raeOffBoresight(rae, sensor, face, maxSensorAz);
    const { u, v } = azel2uv(az, el, sensor.beamwidthRad);
    return { rng: rae.rng, u, v };
}
/**
 * Converts U and V to Azimuth and Elevation off of boresight.
 * @param u The U coordinate.
 * @param v The V coordinate.
 * @param coneHalfAngle The cone half angle of the radar.
 * @returns Azimuth and Elevation off of boresight.
 */
export function uv2azel(u, v, coneHalfAngle) {
    if (u > 1 || u < -1) {
        throw new RangeError(`u is out of bounds: ${u}`);
    }
    if (v > 1 || v < -1) {
        throw new RangeError(`v is out of bounds: ${v}`);
    }
    const alpha = Math.asin(u);
    const beta = Math.asin(v);
    const az = ((alpha / 90) * (coneHalfAngle * RAD2DEG));
    const el = ((beta / 90) * (coneHalfAngle * RAD2DEG));
    return { az, el };
}
/**
 * Converts coordinates from East-North-Up (ENU) to Right-Front-Up (RF) coordinate system.
 * @param enu - The ENU coordinates to be converted.
 * @param enu.x - The east coordinate.
 * @param enu.y - The north coordinate.
 * @param enu.z - The up coordinate.
 * @param az - The azimuth angle in radians.
 * @param el - The elevation angle in radians.
 * @returns The converted RF coordinates.
 */
export function enu2rf({ x, y, z }, az, el) {
    const xrf = Math.cos(el) * Math.cos(az) * x - Math.sin(az) * y + Math.sin(el) * Math.cos(az) * z;
    const yrf = Math.cos(el) * Math.sin(az) * x + Math.cos(az) * y + Math.sin(el) * Math.sin(az) * z;
    const zrf = -Math.sin(el) * x + Math.cos(el) * z;
    return {
        x: xrf,
        y: yrf,
        z: zrf,
    };
}
//# sourceMappingURL=transforms.js.map